#!/usr/bin/env bash

set -eo pipefail

STARTS_WITH_NUMBER='[0-9]+'
CONTINUES_WITH_DASH='([0-9]+)\-'
HAS_DOT='[0-9]+\.'

DIR="${BASH_SOURCE%/*}"
if [[ ! -d "$DIR" ]]; then DIR="$PWD"; fi
# shellcheck source=lib/common.sh
. "$DIR/lib/common.sh"
# shellcheck source=lib/git.sh
. "$DIR/lib/git.sh"

function usage() {
  cat <<EOF

Usage: xivocc-build.sh <release_type>
where <release_type> is one of
  - dev
        Build & publish a dev docker image (used by jenkins)
  - rc
        Build & publish a RC docker image (used by jenkins)
  - prod
        Docker tag a final docker image based on a previous RC (used by jenkins)
  - tag-rc <version> <branch>
        git tag the project with the given <version> using the given <branch>.
        - <version> should not include '-rc' and should match TARGET_VERSION
        - <branch> should be a valid git branch
  - tag-prod <version>
        git tag the project with the given <version> final release based on
        the previously tagged rc <version>-rc
  - set-version <version> [branch]
  		set target version to given <version> in given git [branch] or "master"
	  	if not set.
		- <version> should be a non existing version with correct formatting
		- [branch] should be a valid branch, "master" will be used if not set

EOF
  exit 1
}

function usage_target_version() {
  echo "TARGET_VERSION file is missing or invalid, create one containing:"
  echo "  DOCKER_REPOSITORY=<REPOSITORY>"
  echo "  TARGET_VERSION=<VERSION>"
  echo "where "
  echo " - <REPOSITORY> is the docker repository (i.e xivoxc/xucmgt)"
  echo " - <VERSION> is of form <MAJOR>.<MINOR>.<PATCH> (i.e. 2017.03.02)"
  exit 1
}

function check_git_tag() {
  local EXPECTED_TAG="${1}"
  shift
  local GIT_VERSION
  GIT_VERSION=$(git_get_current_tag)
  if [ -z "$GIT_VERSION" ] || [ ! "${GIT_VERSION}" == "${EXPECTED_TAG}" ]; then
    echo "Git version (${GIT_VERSION}) and target version (${EXPECTED_TAG}) mismatch."
    echo "Check the content of the TARGET_VERSION file matches the git tag"
    exit 1
  fi
}

function ensure_git_tag_inexistent() {
  local tag="${1}"
  shift
  if git_tag_exists "$tag"; then
    echo "$tag is already tagged, cannot continue."
    echo "Did you forget to increment minor or patch number ?"
    exit 1
  fi
}

function ensure_git_tag_exists() {
  local tag="${1}"
  shift
  if ! git_tag_exists "${tag}"; then
    echo "${tag} doesn't exist, cannot continue."
    exit 1
  fi
}

function docker_tag_push() {
  echo "Tagging ${DOCKER_REPOSITORY}:${DOCKER_TAG}"
  if ! docker tag "${DOCKER_REPOSITORY}:${TARGET_VERSION}" "${DOCKER_REPOSITORY}:${DOCKER_TAG}"; then
    echo "Error while tagging docker ${DOCKER_REPOSITORY}:${TARGET_VERSION} to ${DOCKER_REPOSITORY}:${DOCKER_TAG}"
    exit 1
  fi
  if ! docker push "${DOCKER_REPOSITORY}:${DOCKER_TAG}"; then
    echo "Error while pushing ${DOCKER_REPOSITORY}:${DOCKER_TAG}"
    exit 1
  fi
}

function build() {
  echo "Building $TARGET_VERSION (semver ${TARGET_VERSION_SEMVER})"
  export TARGET_VERSION TARGET_VERSION_SEMVER DOCKER_REPOSITORY DOCKER_TAG
  if bash ./docker_build.sh "${TARGET_VERSION}"; then
    docker_tag_push
  else
    echo "docker_build.sh failed, exiting"
    exit 1
  fi
}

function load_target_version_file() {

  if [ ! -f ./TARGET_VERSION ]; then
    echo "Doesn't exist"
    usage_target_version
  fi

  source ./TARGET_VERSION

  if [ -z "$TARGET_VERSION" ]; then
    echo "empty"
    usage_target_version
  fi

  if [ -z "$DOCKER_REPOSITORY" ]; then
    echo "empty repo"
    usage_target_version
  fi

  if [ ! -f ./docker_build.sh ]; then
    echo "docker_build.sh not found !"
    exit 1
  fi

  # Check version, split in $VARRAY
  split "${TARGET_VERSION}" "." VARRAY
  for i in {0..2}; do
    if [ -z "${VARRAY[$i]}" ]; then
      echo "invalid varray: ${VARRAY[1]}"
      usage_target_version
    fi
  done
  TARGET_VERSION_SEMVER=${VARRAY[0]#0}.${VARRAY[1]#0}.${VARRAY[2]#0}
  DOCKER_BRANCH="${VARRAY[0]}.${VARRAY[1]}"
}

function write_target_version_file() {

	ARG_TARGET_VERSION="${1}"
	shift
	ARG_DOCKER_REPOSITORY="${1}"
	shift

	
	if [ -z "$ARG_TARGET_VERSION" ]; then
		echo "empty version"
		exit 1
	fi

	if [ -z "$ARG_DOCKER_REPOSITORY" ]; then
		echo "empty repo"
		exit 1
	fi


  # Check version, split in $VARRAY
  split "${ARG_TARGET_VERSION}" "." VARRAY
  for i in {0..2}; do
    if [ -z "${VARRAY[$i]}" ]; then
      echo "invalid varray: ${VARRAY[1]}"
      usage_target_version
    fi
  done

  echo "DOCKER_REPOSITORY=${ARG_DOCKER_REPOSITORY}" > ./TARGET_VERSION
  echo "TARGET_VERSION=${ARG_TARGET_VERSION}" >> ./TARGET_VERSION
}

function set_docker_tag() {
  GIT_BRANCH="${1}"

  if [[ $GIT_BRANCH =~ $STARTS_WITH_NUMBER ]]; then
    if [[ $GIT_BRANCH =~ $CONTINUES_WITH_DASH ]]; then
      TICKET_NUMBER="${BASH_REMATCH[1]}"
    fi
    if [ "${TICKET_NUMBER}" == "" ]; then
      DOCKER_TAG="${DOCKER_BRANCH}.latestdev"
    else
      DOCKER_TAG="${DOCKER_BRANCH}.dev-${TICKET_NUMBER}"
    fi
  else
    DOCKER_TAG="${DOCKER_BRANCH}.latestdev"
  fi
}

function validate_dev_branch_name() {
  GIT_BRANCH="${1}"
  if [[ $GIT_BRANCH =~ $STARTS_WITH_NUMBER ]]; then
    if [[ $GIT_BRANCH =~ $CONTINUES_WITH_DASH ]]; then
      echo "Valid dev branch ${GIT_BRANCH}"
    elif [[ $GIT_BRANCH =~ $HAS_DOT ]]; then
      echo "Valid LTS dev branch ${GIT_BRANCH}"
    else
      echo "Invalid dev branch ${GIT_BRANCH}"
      exit 1
    fi
  else
    echo "Valid dev branch ${GIT_BRANCH}"
  fi
}

function build_dev() {
  GIT_BRANCH="${1}"


  load_target_version_file
  ensure_git_tag_inexistent "${TARGET_VERSION}"

  COMMIT=$(git_get_commit_id)

  validate_dev_branch_name "${GIT_BRANCH}"
  set_docker_tag "${GIT_BRANCH}"

  TARGET_VERSION="${TARGET_VERSION}-dev-${COMMIT}"
  TARGET_VERSION_SEMVER="${TARGET_VERSION_SEMVER}-dev-${COMMIT}"
  build
}

function build_rc() {
  load_target_version_file
  DOCKER_TAG="${DOCKER_BRANCH}.latestrc"
  ensure_git_tag_inexistent "${TARGET_VERSION}-rc"
  ensure_git_tag_inexistent "${TARGET_VERSION}"
  build
  if [ ! "$SKIP_GIT_TAG" = true ]; then
        tag_rc "${TARGET_VERSION}" "${BRANCH_OR_TAG}"
        ensure_git_tag_exists "${TARGET_VERSION}-rc"
  fi
}

function promote_prod() {
  load_target_version_file
  ensure_git_tag_inexistent "${TARGET_VERSION}"
  if [ ! "$SKIP_GIT_TAG" = true ]; then
        tag_prod "${TARGET_VERSION}" "${BRANCH_OR_TAG}"
        check_git_tag "${TARGET_VERSION}"
  fi
  DOCKER_TAG="${DOCKER_BRANCH}.latest"
  if docker inspect "${DOCKER_REPOSITORY}:${TARGET_VERSION}" >/dev/null 2>&1; then
    docker_tag_push
    # Push exact version for production only
    if ! docker push "${DOCKER_REPOSITORY}:${TARGET_VERSION}"; then
      echo "Failed to push ${DOCKER_REPOSITORY}:${TARGET_VERSION} failed !"
      exit 1
    fi
  else
    echo "Docker image for ${TARGET_VERSION} not found"
    exit 1
  fi
}

function tag_rc() {
  ARG_VERSION="${1}"
  shift
  GIT_BRANCH="${1}"
  shift

  if [ -z "$ARG_VERSION" ]; then
    printf "\t<version> argument missing\n"
    usage
    exit 1
  fi
  if [ -z "$GIT_BRANCH" ]; then
    printf "\t<branch> argument missing\n"
    usage
    exit 1
  fi
  echo "~~ Checking out $GIT_BRANCH ~~"
  git checkout "$GIT_BRANCH"
  echo "~~ Cleaning git workspace ~~"
  git reset --hard
  git clean -f -d
  echo "~~ Pulling latest changes from server ~~"
  git pull >/dev/null
  GIT_IS_CLEAN=$(git status --porcelain)
  if [ -n "$GIT_IS_CLEAN" ]; then
    echo "Your workspace is not clean !"
    exit 1
  fi
  ensure_git_tag_inexistent "${ARG_VERSION}-rc"

  load_target_version_file
  if [ -z "$ARG_VERSION" ] || [ ! "${ARG_VERSION}" == "${TARGET_VERSION}" ]; then
    echo "Argument version (${ARG_VERSION}) and target version (${TARGET_VERSION}) mismatch."
    echo "Check the content of the TARGET_VERSION file"
    exit 1
  fi

  echo "~~ Tagging ${TARGET_VERSION}-rc ~~"
  git tag -a "${TARGET_VERSION}-rc" -m "${TARGET_VERSION} Release candidate"
  echo "~~ Pushing ${TARGET_VERSION}-rc to origin ~~"
  git push origin "${TARGET_VERSION}-rc"
}

function tag_prod() {
  ARG_VERSION="${1}"
  shift
  GIT_BRANCH="${1}"
  shift
  if [ -z "$ARG_VERSION" ]; then
    printf "\t<version> argument missing\n"
    usage
    exit 1
  fi

  echo "~~ Checking out $GIT_BRANCH ~~"
  git checkout "$GIT_BRANCH"

  ensure_git_tag_exists "${ARG_VERSION}-rc"
  load_target_version_file
  ensure_git_tag_inexistent "${TARGET_VERSION}"

  git tag -a "${ARG_VERSION}" -m "${TARGET_VERSION} Final release" "${ARG_VERSION}-rc"
  git push origin "${ARG_VERSION}"
}

function set_version() {
  ARG_VERSION="${1}"
  shift
  GIT_BRANCH="${1}"
  shift

  if [ -z "$ARG_VERSION" ]; then
    printf "\t<version> argument missing\n"
    usage
    exit 1
  fi
  if [ -z "$GIT_BRANCH" ]; then
    printf "\t<branch> argument missing, using 'master' instead \n"
	GIT_BRANCH="master"
  fi
  echo "~~ Checking out $GIT_BRANCH ~~"
  git checkout "$GIT_BRANCH"
  echo "~~ Cleaning git workspace ~~"
  git reset --hard
  git clean -f -d
  echo "~~ Pulling latest changes from server ~~"
  git pull >/dev/null
  GIT_IS_CLEAN=$(git status --porcelain)
  if [ -n "$GIT_IS_CLEAN" ]; then
    echo "Your workspace is not clean !"
    exit 1
  fi

  ensure_git_tag_inexistent "${ARG_VERSION}"
  load_target_version_file
  if [ "$ARG_VERSION" = "$TARGET_VERSION" ]; then
	  echo "Target version is already set to $ARG_VERSION"
	  exit 0
  else
	  write_target_version_file "$ARG_VERSION" "$DOCKER_REPOSITORY"

	  git add TARGET_VERSION
	  git commit -m "Bump to ${ARG_VERSION}"
	  git push origin
  fi
}

if [ -z "$1" ]; then
  echo "Required argument missing: <release_type>"
  usage
fi

if options=$(getopt -l "skip-tag" -o "s" -a -- "$@"); then
  case "${options[@]}" in
    *-s*|*--skip-tag*)
      echo "Git tagging will be skipped"
      SKIP_GIT_TAG=true
      ;;
  esac
fi


case $1 in
    dev)
		  build_dev "$2"
		  ;;

    rc)
		  build_rc
      ;;

    prod)
      promote_prod
      ;;

    tag-rc)
      tag_rc "$2" "$3"
      ;;

	  tag-prod)
      tag_prod "$2" "$3"
		  ;;
	
    set-version)
		  set_version "$2" "$3"
		  ;;
    *)
      echo "Invalid <release_type>: $1"
      usage
      ;;
esac
